﻿using System;
using System.Net.Sockets;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;
using System.Threading;

namespace MockSmtpServer
{
    public class SmtpServer
    {
        private const string replyOK = "250 OK";
        private const string replySyntax = "500 Syntax error, command unrecognized";
        private const string replyBadArgs = "501 Syntax error in parameters or arguments";


        private string domain;

        public int Port { get; set; }

        public event EventHandler<SmtpMessageReceivedEventArgs> MessageReceived;

        public SmtpServer()
            : this(25)
        {
            this.domain = Environment.MachineName;
        }

        public SmtpServer(int port)
        {
            this.Port = port;

            Thread thread = new Thread(new ThreadStart(MailHandler));
            thread.IsBackground = true;
            thread.Start();
        }

        private void MailHandler()
        {
            TcpListener server = new TcpListener(IPAddress.Any, Port);
            server.Start();

            while (true)
            {
                TcpClient client = server.AcceptTcpClient();

                NetworkStream stream = client.GetStream();
                StreamReader reader = new StreamReader(stream, Encoding.ASCII, false);
                StreamWriter writer = new StreamWriter(stream, Encoding.ASCII);

                WriteResponse(writer, "220 {0} Service ready", domain);

                bool quit = false;
                string command;
                SmtpMailMessage message = new SmtpMailMessage();

                while (client.Connected && !quit && (command = reader.ReadLine()) != null)
                {
                    command = command.Trim();
                    if (command.Length >= 4)
                    {
                        switch (command.Substring(0, 4).ToUpper())
                        {
                            case "HELO":
                                WriteResponse(writer, "250 {0}", IPAddress.Parse(((IPEndPoint)client.Client.RemoteEndPoint).Address.ToString()));
                                break;
                            case "MAIL":
                                message = new SmtpMailMessage();
                                if (command.Length > 10 && command.Substring(4, 6).ToUpper() == " FROM:")
                                {
                                    message.From = command.Substring(10);
                                    WriteResponse(writer, replyOK);
                                }
                                else
                                {
                                    WriteResponse(writer, replyBadArgs);
                                }
                                break;
                            case "RCPT":
                                if (command.Length > 8 && command.Substring(4, 4).ToUpper() == " TO:")
                                {
                                    message.AddRecipient(command.Substring(8));
                                    WriteResponse(writer, replyOK);
                                }
                                else
                                {
                                    WriteResponse(writer, replyBadArgs);
                                }
                                break;
                            case "DATA":
                                WriteResponse(writer, "354 Start mail input; end with <CRLF>.<CRLF>");
                                string line;
                                while ((line = reader.ReadLine()) != null && line != ".")
                                {
                                    if (line.Length > 0 && line[0] == '.')
                                    {
                                        line = line.Substring(1);
                                    }
                                    message.AddMessageLine(line);
                                }
                                OnMessageReceived(message);
                                WriteResponse(writer, replyOK);
                                break;
                            case "RSET":
                                message = new SmtpMailMessage();
                                WriteResponse(writer, replyOK);
                                break;
                            case "VRFY":
                            // To support RFC 5321 and since this is a mock, all addresses are valid 
                            case "NOOP":
                                WriteResponse(writer, replyOK);
                                break;
                            case "QUIT":
                                WriteResponse(writer, "221 {0} Service closing transmission channel", domain);
                                quit = true;
                                break;
                            default:
                                // We only implement the minimum command set
                                WriteResponse(writer, replySyntax);
                                break;
                        }
                    }
                    else
                    {
                        // Reply to garbage with
                        WriteResponse(writer, replySyntax);
                    }
                }

                client.Close();
            }
        }

        private void WriteResponse(StreamWriter writer, string format, params object[] arg)
        {
            writer.WriteLine(format, arg);
            writer.Flush();
        }

        private void WriteResponse(StreamWriter writer, string response)
        {
            writer.WriteLine(response);
            writer.Flush();
        }

        public void OnMessageReceived(SmtpMailMessage message)
        {
            EventHandler<SmtpMessageReceivedEventArgs> handler = MessageReceived;

            if (handler != null)
            {
                handler(this, new SmtpMessageReceivedEventArgs(message));
            }
        }
    }

    public class SmtpMessageReceivedEventArgs : EventArgs
    {
        public SmtpMailMessage Message { get; set; }

        public SmtpMessageReceivedEventArgs(SmtpMailMessage message)
        {
            this.Message = message;
        }
    }
}
